-- rsiTurtle.lua
-- Created on:  2020-10-16
-- Author:      Rétro
--
-- Rétro Stonks Industries - Tout droits réservés
--
-- API pour tortues RSI

---
-- \class Tortue RSI
local RSI_Turtle = {}
RSI_Turtle.__index = RSI_Turtle

require("printf")
require("macros")

---
-- \brief Affiche un hello world
function RSI_Turtle:helloWorld()
    print("Hello World!")
end

---
-- \brief Constructeur de RSI_Turtle
function RSI_Turtle:init()
    local turtle = {}
    setmetatable(turtle,RSI_Turtle)
    
    ---
    -- \brief Dictionnaire pour la position (x,y,z).
    --        Le système de coordonnées est celui de minecraft:
    --          x: Avant
    --          y: Altitude
    --          z: Droite
    turtle.pos_ = {}
    turtle.pos_["x"] = 0
    turtle.pos_["y"] = 0
    turtle.pos_["z"] = 0
    
    ---
    -- \brief Orientation de la tortue en cadrant horloge:
    --          0:  Avant
    --          3:  Droite
    --          6:  Arrière
    --          9:  Gauche
    turtle.facing_ = 0
    
    return turtle
end


---
-- \brief Affiche la position et l'orientation de la tortue
function RSI_Turtle:print()
    printf("Position:\n X=%d\n Y=%d\n Z=%d\n", self.pos_["x"],
                    self.pos_["y"], self.pos_["z"])
    printf("Orientation: ")

    if( self.facing_ == 3 )
    then
        print("Droite")
    elseif ( self.facing_ == 6 )
    then
        print("Arriere")
    elseif ( self.facing_ == 9 )
    then
        print("Gauche")
    elseif ( self.facing_ == 0 )
    then
        print("Avant")
    end
end


---
-- \brief   Affiche la position et l'orientation de la tortue
-- \return int,int,int  Positions X, Y et Z
function RSI_Turtle:getPos()
    return self.pos_["x"], self.pos_["y"], self.pos_["z"]
end


---
-- \brief Déplacement parallèle au sol
-- \return bool Succès
function RSI_Turtle:moveTo2D(x, z)
    return self:move2D(x - self.pos_["x"], z - self.pos_["z"])
end


---
-- \brief Déplacement dans les trois dimensions
-- \return bool Succès
function RSI_Turtle:moveTo(x, y, z)
    return self:move(x - self.pos_["x"], y - self.pos_["y"], z - self.pos_["z"])
end


---
-- \brief Retourner au point d'origine (0,0,0)
-- \return bool Succès
function RSI_Turtle:moveToOrigin()
    return self:moveTo(0,0,0)
end


---
-- \brief Retourner au point d'origine (0,0) tout en restant au niveau du sol
-- \return bool Succès
function RSI_Turtle:moveToOriginGrounded()
    return self:moveToGrounded(0,0)
end


---
-- \brief Déplacement parallèle au sol
--        Ordre de déplacement: X, Z
-- \param int       dx          Déplacement vers l'avant (+) ou l'arrière (-)
-- \param int       dz          Déplacement vers la gauche (-) ou la droite (+)
-- \param bool      grounded    Se déplacer en restant toujours au niveau du sol
-- \return bool Succès
function RSI_Turtle:move2D(dx, dz, grounded)
    local success = true

    if(grounded)
    then
        self:ground()
    end

    if dx >= 0
    then
        self:face(12)
        if(grounded)
        then
            success = self:moveForwardGrounded(dx)
        else
            success = self:moveForward(dx)
        end
    elseif dx < 0
    then
        self:face(6)
        if(grounded)
        then
            success = self:moveForwardGrounded(-1*dx)
        else
            success = self:moveForward(-1*dx)
        end
    end

    if not success
    then
        return false
    end
    
    if (self.facing_ == 0)
    then
        if dz < 0
        then
            self:turnLeft()
        elseif dz > 0
        then
            self:turnRight()
        end
    elseif (self.facing_ == 6)
    then
        if dz > 0
        then
            self:turnLeft()
        elseif dz < 0
        then
            self:turnRight()
        end
    end
    
    if not success
    then
        return false
    end

    if(grounded)
    then
        success = self:moveForwardGrounded(math.abs(dz))
    else
        success = self:moveForward(math.abs(dz))
    end

    return success
end


---
-- \brief Déplacement dans les trois dimensions
--        Ordre de déplacement: X, Z, Y
-- \return bool Succès
function RSI_Turtle:move(dx, dy, dz)
    local success = true

    if (dy > 0)
    then
        for i=1,dy,1
        do
            success = self:up()
        end
    elseif dy < 0
    then
        for i=-1,dy,-1
        do
            success = self:down()
        end
    end

    if (success) then
        return self:move2D(dx, dz)
    else
        return false
    end
end


---
-- \brief Déplacement avant multiple
-- \param distance Distance à parcourir
-- \return bool Succès
function RSI_Turtle:moveForwardGrounded(distance, action)
    return self:moveForward(distance, true, action)
end


---
-- \brief Déplacement s'assurant de rester au niveau du sol
-- escalade et descend automatiquement les pentes
-- \param int       dx          Déplacement vers l'avant (+) ou l'arrière (-)
-- \param int       dz          Déplacement vers la gauche (-) ou la droite (+)
-- \return bool Succès
function RSI_Turtle:moveToGrounded(x, z)
    return self:moveGrounded(x - self.pos_["x"], z - self.pos_["z"])
end


---
-- \brief Déplacement s'assurant de rester au niveau du sol
-- escalade et descend automatiquement les pentes
-- \param int       dx          Déplacement vers l'avant (+) ou l'arrière (-)
-- \param int       dz          Déplacement vers la gauche (-) ou la droite (+)
-- \return bool Succès
function RSI_Turtle:moveGrounded(dx, dz)
    return self:move2D(dx, dz, true)
end


---
-- \brief Tourne la tortue face à la direction voulue.
-- \param Orientation en heures (3, 6, 9, 12)
-- \return bool Succès
function RSI_Turtle:face(orientation)
    if (orientation ~= 3 and orientation ~= 6 and 
        orientation ~= 9 and orientation ~= 12)
    then
        return false
    end

    local diff = (orientation - self.facing_) % 12
    if diff == 9 then
        self:turnLeft()
    elseif diff == 3 then
        self:turnRight()
    elseif diff == 6 then
        self:turnLeft()
        self:turnLeft()
    end
    return true
end


---
-- \brief Tourner la tortue de 180 degrés (par la gauche)
function RSI_Turtle:turn180()
    self:turnLeft();
    self:turnLeft();
end


---
-- \brief Déplacement avant
-- \param bool      grounded    Se déplacer en restant toujours au niveau du sol
-- \return bool Succès
function RSI_Turtle:forward(grounded)
    return self:moveForward(1, grounded)
end

---
-- \brief Déplacement avant multiple
-- \param distance  Distance à parcourir
-- \param bool              grounded    Se déplacer en restant toujours au niveau du sol
-- \param function_pointer  action      Exécuté après chaque déplacement
-- \return bool Succès
function RSI_Turtle:moveForward(distance, grounded, action)
    local traveled = 0

    if (grounded)
    then
        self:ground()
        self:climb()
    end

    while (traveled < distance) and turtle.forward()
    do
        if (action ~= nil)
        then
            action()
        end

        if (grounded)
        then
            self:ground()
            self:climb()
        end

        traveled = traveled + 1
    end

    if self.facing_ == 0 then
        self.pos_["x"] = self.pos_["x"] + traveled
    elseif self.facing_ == 3 then
        self.pos_["z"] = self.pos_["z"] + traveled 
    elseif self.facing_ == 6 then
        self.pos_["x"] = self.pos_["x"] - traveled
    elseif self.facing_ == 9 then
        self.pos_["z"] = self.pos_["z"] - traveled
    end

    return (traveled == distance)
end


---
-- \brief Déplacement arrière
-- \return bool Succès
function RSI_Turtle:backward()
    return self:moveBackward(1)
end


---
-- \brief Déplacement arrière
-- \param distance Distance à parcourir
-- \param bool              grounded    Se déplacer en restant toujours au niveau du sol
-- \param function_pointer  action      Exécuté après chaque déplacement
-- \return bool Succès
function RSI_Turtle:moveBackward(distance, grounded, action)
    local traveled = 0

    if (grounded)
    then
        self:ground()
        self:climb()
    end

    while (traveled < distance) and turtle.back()
    do
        if (action ~= nil)
        then
            action()
        end

        if (grounded)
        then
            self:ground()
            self:climb()
        end

        traveled = traveled + 1
    end
    
    if result then
        if self.facing_ == 0 then
            self.pos_["x"] = self.pos_["x"] - traveled
        elseif self.facing_ == 3 then
            self.pos_["z"] = self.pos_["z"] - traveled
        elseif self.facing_ == 6 then
            self.pos_["x"] = self.pos_["x"] + traveled
        elseif self.facing_ == 9 then
            self.pos_["z"] = self.pos_["z"] + traveled
        end
    end

    return (traveled == distance)
end


---
-- \brief Déplacement haut
-- \return bool Succès
function RSI_Turtle:up()
    result = turtle.up()
    
    if result then
        self.pos_["y"] = self.pos_["y"] + 1
    end
    return result
end


---
-- \brief Déplacement bas
-- \return bool Succès
function RSI_Turtle:down()
    result = turtle.down()
    
    if result then
        self.pos_["y"] = self.pos_["y"] - 1
    end
    return result
end


---
-- \brief Tourne 90 degrés dans le sens contraire des aiguilles d'une montre
function RSI_Turtle:turnLeft()
    turtle.turnLeft()
    self.facing_ = (self.facing_ - 3) % 12
end


---
-- \brief Tourne 90 degrés dans le sens des aiguilles d'une montre
function RSI_Turtle:turnRight()
    turtle.turnRight()
    self.facing_ = (self.facing_ + 3) % 12
end


---
-- \brief Déplacement à gauche
-- \param bool      grounded    Se déplacer en restant toujours au niveau du sol
-- \return bool Succès
function RSI_Turtle:left(grounded)
    self:turnLeft()
    result = self:forward(grounded)
    return result
end


---
-- \brief Déplacement à droite
-- \param bool      grounded    Se déplacer en restant toujours au niveau du sol
-- \return bool Succès
function RSI_Turtle:right(grounded)
    self:turnRight()
    result = self:forward(grounded)
    return result
end

---
-- \brief Augmenter en altitude jusqu'à ce que le bloc en face soit de l'air
-- \return bool Succès
function RSI_Turtle:climb()
    local success = true

    while (not self:isAir("front") and success)
    do
        success = self:up()
    end
    return success
end


---
-- \brief Descendre jusqu'à rencontrer le sol
-- \return bool Succès si a descendu d'au moins un bloc
function RSI_Turtle:ground()
    success = self:down()
    while self:down() do end
    return success
end


---
-- \brief Quadriller une zone carrée. La tortue traverse chaque case, ligne par 
--          ligne. N'affecte pas l'altitude.
-- \param pointeur_de_fonction  action  Pointeur de l'action à réaliser juste
--          après être arrivé sur chaque nouveau bloc.
-- \param int               lines       Nombre de lignes    (direction: avant)
-- \param int               cols        Nombre de colonnes  (direction: droite)
-- \param function_pointer  action      Action immédiatement après chaque déplacement
-- \param bool              grounded    Se déplacer en restant toujours au niveau du sol
-- \return bool Succès
function RSI_Turtle:goThroughArea(lines, cols, action, grounded)
    local success = true

    assert(lines >= 1)
    assert(cols >= 1)

    -- La ligne de la case de départ compte comme 1
    lines = lines - 1

    if (grounded)
    then
        self:ground()
    end
    action()

    success = self:moveForward(lines, grounded, action)
    for i=2,cols,1
    do
        print("Debut de colonne.")

        if ( (i%2) == 1 )
        then
            success = self:left(grounded)
            self:turnLeft()
        else
            success = self:right(grounded)
            self:turnRight()
        end
        action()

        success = self:moveForward(lines, grounded, action)

        if not success
        then
            return success
        end

    end
    return success
end


---
-- \brief Quadriller une zone carrée. La tortue traverse chaque case, ligne par 
--          ligne. Reste au niveau du sol.
-- \param pointeur_de_fonction action Pointeur de l'action à réaliser juste
--          après être arrivé sur chaque nouveau bloc.
-- \return bool Succès
function RSI_Turtle:goThroughAreaGrounded(lines, cols, action)
    return self:goThroughArea(lines, cols, action, true)
end


---
-- \brief Vérifier si le bloc dans la direction en argument est de l'air
-- \param Côté: "top", "bottom", "left", "right", "front", "back"
-- \return bool Bloc en face identique à l'argument
function RSI_Turtle:isAir(side)
    if side == "top"
    then
        return (not turtle.inspectUp())
    elseif side == "down"
    then
        return (not turtle.inspectDown())
    elseif side == "front"
    then
        return (not turtle.inspect())
    elseif side == "back"
    then
        self:turn180()
        return (not turtle.inspect())
    elseif side == "left"
    then
        self:turnLeft()
        return (not turtle.inspect())
    elseif side == "right"
    then
        self:turnRight()
        return (not turtle.inspect())
    end
end


---
-- \brief Vérifier si l'ID du bloc en face contient la regex
-- \param string    blockIdRegex   Regex correspondant au bloc à vérifier
-- \param string    side           Direction parmi: front, back, up, down, left, right
--                                 Si vide, prend "front" par défaut
-- \return bool Bloc en face match la regex
function RSI_Turtle:facesBlock(blockIdRegex, side)
    local isNotAir, blockData

    if side == nil
    then
        side = "front"
    end

    if side == "up"
    then
        isNotAir, blockData = turtle.inspectUp()
    elseif side == "down"
    then
        isNotAir, blockData = turtle.inspectDown()
    elseif side == "front"
    then
        isNotAir, blockData = turtle.inspect()
    elseif side == "back"
    then
        self:turn180()
        isNotAir, blockData = turtle.inspect()
    elseif side == "left"
    then
        self:turnLeft()
        isNotAir, blockData = turtle.inspect()
    elseif side == "right"
    then
        self:turnRight()
        isNotAir, blockData = turtle.inspect()
    end

    if (isNotAir)
    then
        return string.match(blockData.name, blockIdRegex)
    else
        return false
    end
end


---
-- \brief Casser un bloc dans la direction souhaitée
-- \param string    blockIdRegex    Regex correspondant au bloc à casser.
--                                  Si vide, casse n'importe quoi.
-- \param string    side            Direction parmi: front, back, up, down, left, right
--                                  Si vide, prend "front" par défaut
-- \return bool Succès
function RSI_Turtle:dig(blockIdRegex, side)
    --TODO: Tester
    local success

    if blockIdRegex == nil
    then
        blockIdRegex = ''
    end

    if side == nil
    then
        side = "front"
    end

    if (self:facesBlock(blockIdRegex, side))
    then
        if (side == "up")
        then
            success = turtle.digUp()
        elseif (side == "down")
        then
            success = turtle.digDown()
        else
            success = turtle.dig()
        end
    end

    return success
end


---
-- \brief Quadriller une zone carrée. La tortue traverse chaque case, ligne par 
--          ligne. N'affecte pas l'altitude.
-- \param pointeur_de_fonction  action  Pointeur de l'action à réaliser juste
--          après être arrivé sur chaque nouveau bloc.
-- \param int               lines       Nombre de lignes    (direction: avant)
-- \param int               cols        Nombre de colonnes  (direction: droite)
-- \param int               depth       Profondeur          (direction: bas)
-- \param function_pointer  action      Action immédiatement après chaque déplacement
-- \return bool Succès
function RSI_Turtle:goThroughVolume(lines, cols, depth, action)
    local success = true
    
    assert(lines >= 1)
    assert(cols >= 1)
    assert(depth >= 1)

    for i=1,depth,1
    do
        success = self:goThroughArea(lines, cols, action)
        success = self:moveTo2D(0,0)
        action()
        success = self:down()
        self:face(12)
    end
    
    return success
end


---
-- \brief Vérifier si l'ID du bloc en face contient la regex
-- \param int   fuelLevel       Brûler du carburant jusqu'à ce que le tortue
--                              atteigne le niveau spécifié
-- \param int   fuelChestSlot   Case de l'inventaire contenant un ender chest
--                              contenant du carburant
-- \return bool Succès
function RSI_Turtle:refuelUntil(fuelLevel, fuelChestSlot)
    local success
    local itemInFuelSlot

    self:dig() -- Faire de la place pour le coffre
    turtle.select(fuelChestSlot)
    success = turtle.place()
    if (not success) then return false end

    while (turtle.getFuelLevel() < fuelLevel)
    do
        -- Prendre un object dans le coffre
        success = turtle.suck(1)
        if (not success) then return false end

        -- Le brûler
        success = turtle.refuel(1)
        if (not success) then return false end
    end

    -- Ramasser le coffre
    turtle.select(fuelChestSlot)
    success = self:dig()
    if (not success) then return false end

    -- Vérifier qu'on a bien ramassé le coffre au bon emplacement de l'inventaire
    itemInFuelSlot = turtle.getItemDetail(fuelChestSlot)
    if (itemInFuelSlot ~= nil)
    then
        return string.match(itemInFuelSlot.name, ID_ENDERCHEST)
    end
    return false
end


---
-- \brief   Place le contenu de l'inventaire dans un ender chest devant soit
--          (ou le balance par terre s'il n'y a aucun bloc de stockage)
--          Agit sur un intervalle de cases de l'inventaire
--          ex: 1,4 va prendre les cases d'inventaire 1, 2, 3, et 4
-- \param int   firstSlot   Première case
-- \param int   lastSlot    Dernière case
-- \return bool Succès
function RSI_Turtle:dumpInventory(firstSlot, lastSlot, dumpChestSlot)
    local success

    self:dig() -- Faire de la place pour le coffre
    turtle.select(dumpChestSlot)
    success = turtle.place()
    if (not success) then return false end
    
    for i=firstSlot, lastSlot, 1
    do
        turtle.select(i)
        success = turtle.drop() or (turtle.getItemCount() == 0)
        if (not success) then return false end
    end
    
    -- Ramasser le coffre
    turtle.select(dumpChestSlot)
    success = self:dig()
    if (not success) then return false end

    -- Vérifier qu'on a bien ramassé le coffre au bon emplacement de l'inventaire
    itemInDumpChestSlot = turtle.getItemDetail(dumpChestSlot)
    if (itemInDumpChestSlot ~= nil)
    then
        return string.match(itemInDumpChestSlot.name, ID_ENDERCHEST)
    end
    return false
end


---
-- \brief   L'inventaire est plein (aucune case totalement vide) entre les
--          cases données
--          Agit sur un intervalle de cases de l'inventaire
--          ex: 1,4 va prendre les cases d'inventaire 1, 2, 3, et 4
-- \param int   firstSlot   Première case
-- \param int   lastSlot    Dernière case
-- \return bool Inventaire plein
function RSI_Turtle:inventoryFull(firstSlot, lastSlot)
    for i=firstSlot, lastSlot, 1
    do
        turtle.select(i)
        if (turtle.getItemCount() == 0)
        then
            return false
        end
    end
    return true
end


return RSI_Turtle
